#include <stdio.h>
#include <stdlib.h>
#include <string.h> /* for strerror() */
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>

#include <sys/types.h>
#include <sys/stat.h>

#include <gtk/gtk.h>

#define DATA_FILE "./exlate.conf"

typedef struct EXULine 
{
  char   iso4217[4];
  char   cur_label[41];
  double rate_s,
         rate_i;
} EXULine;

typedef struct s_EXUList EXUList;
typedef struct s_EXUList
{
	EXULine *data;
	EXUList *prev, 
                *next;
} s_EXUList;

EXUList *Current_Table = NULL;

EXULine *find_by_iso4217(char *TLA)
{
  EXUList *cursor = Current_Table;
  EXULine *ret = NULL;
  
  while(cursor && !ret)
  {
    if(strcmp(TLA,cursor->data->iso4217))
      cursor = cursor->next;
    else
      ret = cursor->data;
  }
  return ret;
}

void delete_table(EXUList **table)
{
  EXUList *temp;
  while(*table)
  {
    if ((*table)->data)
      free((*table)->data);
    
    temp = (*table)->next;
    free(*table);
    *table = temp;
  }
  *table = NULL;
}

EXUList *insert_line(EXUList **list, EXULine * line)
{
  EXUList *cursor, *ins_pos = NULL;
  EXULine *cell = (EXULine *) malloc(sizeof(EXULine));
  short   before = 0; /* non zero if have to insert before ins_pos */

    memcpy(cell,line,sizeof(EXULine)); /* copy data into mem cell */
    cursor = *list;
    
    if(!cursor) /* if empty list (first record, initialize list) */
    {
      cursor = *list = (EXUList*) malloc(sizeof(EXUList));
      cursor->prev = cursor->next = NULL;
    }
    else
    {
      while(cursor && !ins_pos) /* find position to insert line at */
      {
        if(strcmp(cursor->data->iso4217, line->iso4217)>0)
        { ins_pos = cursor; before = 1; }
        else if(cursor->next) cursor = cursor->next;
             else ins_pos = cursor; /* no next record, stop and assertz() */
      }
      cursor = (EXUList*) malloc(sizeof(EXUList));
      
      if(before)
      {
        cursor->next = ins_pos; cursor->prev = ins_pos->prev;
        if(ins_pos == *list) *list = cursor; /* in case it is in first position of the list */
      }
      else { cursor->next = ins_pos->next; cursor->prev = ins_pos; }
      if(cursor->next) cursor->next->prev = cursor;
      if(cursor->prev) cursor->prev->next = cursor;
    }
    cursor->data = cell;
    return cursor;
}

char*ignore_trailing(char *st)
{
  while(*st == ' ') st--;
  return st;
}

double get_trailing_val(char**st)
{
  char *end_cur,
       *str_cur = *st; /* just an alias, to simplify reading */
  double ret;
  char *t_str;
  unsigned short lg;
  str_cur=ignore_trailing(str_cur);
  end_cur = str_cur;
  while(*str_cur != ' ') str_cur--;
  *st = str_cur; /* update position for caller */
  lg = end_cur - str_cur;
  memcpy(t_str = (char*) malloc(lg+1), str_cur + 1, lg);
  t_str[lg] = '\0';
  ret = atof(t_str);
  free(t_str);
  return ret;
}

EXUList *load_table(char * file_name, EXUList **table)
{
  EXULine temp;
  EXUList *ret = NULL;
  char *str_cur, *end_cur, *cpy;
  char *r_line;
  FILE * handle;
  unsigned int cur_pos = 0, old_pos, file_size;
  unsigned char l_line, this_char;
  
  if(*table) delete_table(table);
  
  if((handle = fopen(file_name, "r")) != NULL)
  {
    fseek(handle,0,SEEK_END); file_size = ftell(handle) + 1; fseek(handle,0,SEEK_SET);
    while (cur_pos < file_size)
    {
      do /* find next non-blank line */
      {
        fread(&this_char,1,1,handle);
	cur_pos++;
      } while((cur_pos < file_size) && (this_char < ' '));
      
      if(this_char >= ' ')
      {
        old_pos = ftell(handle) - 1;
        l_line = 0;
        while((cur_pos < file_size) && (this_char >= ' '))
        {
          fread(&this_char,1,1,handle);
	  cur_pos++; l_line++;
        }
        if(l_line > 0)
        {
          r_line = (char*)malloc(l_line + 1);
          fseek(handle, old_pos, SEEK_SET);
          fread(r_line, l_line, 1, handle);
          r_line[l_line]='\0';
	  cur_pos--;
	  if(r_line[0] == '#')
	  {
	    /* process directive */
	    puts(r_line);
	  }
	  else
	  {
            /* process line as currency data */
            memcpy(temp.iso4217,r_line,3); temp.iso4217[3]='\0'; /* copy the ISO4217 code */
            str_cur = r_line;
            end_cur = &r_line[4];
            cpy = temp.cur_label;
            while(str_cur[1]) str_cur++; /* send cursor through the end of string */
            temp.rate_i = get_trailing_val(&str_cur);
            temp.rate_s = get_trailing_val(&str_cur);
	    str_cur = ignore_trailing(str_cur);
            while(end_cur <= str_cur) *(cpy++) = *(end_cur++);
            *cpy='\0';
	    insert_line(table,&temp);
	  }
          free(r_line);
        }
      }
    }
    fclose(handle);
    ret = *table;
  }
  else
  {
    printf("Fatal: %s\n",strerror(errno));
  }
  return ret;
}

/********************************************************************/
/* the GTK interface                                                */
/********************************************************************/

  GtkWidget *about_window;

void about_window_close(GtkWidget *widget, GtkSelectionData *data,
                        guint info, guint time, gpointer user_data)
{
  gtk_widget_destroy(GTK_WIDGET(about_window));
}


void AboutDialog(void)
{
  GtkWidget *vbox1;
  GtkWidget *alignment1;
  GtkWidget *vbox2;
  GtkWidget *frame1;
  GtkWidget *frame2;
  GtkWidget *vbox3;
  GtkWidget *label5;
  GtkWidget *hseparator1;
  GtkWidget *label6;
  GtkWidget *label8;
  GtkWidget *ok_button;

  about_window = gtk_window_new(GTK_WINDOW_DIALOG);
  gtk_widget_set_name(about_window, "about_window");
  gtk_object_set_data(GTK_OBJECT (about_window), "about_window", about_window);
  gtk_window_set_title(GTK_WINDOW(about_window), "About...");
  gtk_window_set_modal(GTK_WINDOW(about_window), TRUE);

  vbox1 = gtk_vbox_new (FALSE, 0);
  gtk_widget_set_name (vbox1, "vbox1");
  gtk_widget_ref (vbox1);
  gtk_object_set_data_full(GTK_OBJECT (about_window), "vbox1", vbox1,
                           (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (vbox1);
  gtk_container_add (GTK_CONTAINER (about_window), vbox1);

  alignment1 = gtk_alignment_new (0.5, 0.5, 1, 1);
  gtk_widget_set_name (alignment1, "alignment1");
  gtk_widget_ref (alignment1);
  gtk_object_set_data_full (GTK_OBJECT (about_window), "alignment1", alignment1,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (alignment1);
  gtk_box_pack_start (GTK_BOX (vbox1), alignment1, TRUE, TRUE, 0);

  vbox2 = gtk_vbox_new (FALSE, 0);
  gtk_widget_set_name (vbox2, "vbox2");
  gtk_widget_ref (vbox2);
  gtk_object_set_data_full (GTK_OBJECT (about_window), "vbox2", vbox2,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (vbox2);
  gtk_container_add (GTK_CONTAINER (alignment1), vbox2);

  frame1 = gtk_frame_new (NULL);
  gtk_widget_set_name (frame1, "frame1");
  gtk_widget_ref (frame1);
  gtk_object_set_data_full (GTK_OBJECT (about_window), "frame1", frame1,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (frame1);
  gtk_box_pack_start (GTK_BOX (vbox2), frame1, TRUE, TRUE, 0);

  frame2 = gtk_frame_new (NULL);
  gtk_widget_set_name (frame2, "frame2");
  gtk_widget_ref (frame2);
  gtk_object_set_data_full (GTK_OBJECT (about_window), "frame2", frame2,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (frame2);
  gtk_container_add (GTK_CONTAINER (frame1), frame2);
  gtk_widget_set_usize (frame2, 248, 56);
  gtk_container_set_border_width (GTK_CONTAINER (frame2), 1);
  gtk_frame_set_shadow_type (GTK_FRAME (frame2), GTK_SHADOW_OUT);

  vbox3 = gtk_vbox_new (FALSE, 0);
  gtk_widget_set_name (vbox3, "vbox3");
  gtk_widget_ref (vbox3);
  gtk_object_set_data_full (GTK_OBJECT (about_window), "vbox3", vbox3,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (vbox3);
  gtk_container_add (GTK_CONTAINER (frame2), vbox3);

  label5 = gtk_label_new ("Currency Converter for Linux / GTK+");
  gtk_widget_set_name (label5, "label5");
  gtk_widget_ref (label5);
  gtk_object_set_data_full (GTK_OBJECT (about_window), "label5", label5,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (label5);
  gtk_box_pack_start (GTK_BOX (vbox3), label5, TRUE, FALSE, 0);

  hseparator1 = gtk_hseparator_new ();
  gtk_widget_set_name (hseparator1, "hseparator1");
  gtk_widget_ref (hseparator1);
  gtk_object_set_data_full (GTK_OBJECT (about_window), "hseparator1", hseparator1,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (hseparator1);
  gtk_box_pack_start (GTK_BOX (vbox3), hseparator1, TRUE, TRUE, 0);

  label6 = gtk_label_new ("By Regis LECLERC 2000 (regis.leclerc@usa.net)");
  gtk_widget_set_name (label6, "label6");
  gtk_widget_ref (label6);
  gtk_object_set_data_full (GTK_OBJECT (about_window), "label6", label6,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (label6);
  gtk_box_pack_start (GTK_BOX (vbox3), label6, TRUE, TRUE, 0);

  label8 = gtk_label_new ("This small application (designed with Glade and written in one"\
                          " night in C with GTK+) aims at giving you a conversion of prices"\
			  " in foreign currencies for the things you buy on the internet.\n\n"\
			  "If you are looking for a skilled C programmer, drop me a mail at"\
			  " regis.leclerc@usa.net\n\n(USA/Australia/Canada employers"\
			  " welcome - sponsorship needed)");
  gtk_widget_set_name (label8, "label8");
  gtk_widget_ref (label8);
  gtk_object_set_data_full (GTK_OBJECT (about_window), "label8", label8,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (label8);
  gtk_box_pack_start (GTK_BOX (vbox2), label8, TRUE, TRUE, 0);
  gtk_widget_set_usize (label8, 400, -2);
  gtk_label_set_justify (GTK_LABEL (label8), GTK_JUSTIFY_LEFT);
  gtk_label_set_line_wrap (GTK_LABEL (label8), TRUE);
  gtk_misc_set_alignment (GTK_MISC (label8), 0.3, 0.5);
  gtk_misc_set_padding (GTK_MISC (label8), 10, 0);

  ok_button = gtk_button_new_with_label ("Gotcha!");
  gtk_widget_set_name (ok_button, "ok_button");
  gtk_widget_ref (GTK_WIDGET(ok_button));
  gtk_object_set_data_full (GTK_OBJECT(about_window), "ok_button", ok_button,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_signal_connect(GTK_OBJECT(ok_button), "clicked", (GtkSignalFunc)about_window_close, NULL);
  gtk_widget_show (GTK_WIDGET(GTK_BUTTON(ok_button)));
  gtk_box_pack_start (GTK_BOX (vbox1), ok_button, FALSE, TRUE, 0);
  gtk_container_set_border_width (GTK_CONTAINER (ok_button), 4);
  gtk_widget_show(GTK_WIDGET(about_window));
}

/* GtkWidgets, declared as global to be accessible from outwards */
  GtkWidget *source_name;
  GtkWidget *source_rate;
  GtkWidget *target_name;
  GtkWidget *target_rate;
  GtkWidget *result_label;
  GtkWidget *status_bar;

  guint status_bar_context;

  EXULine *Source_Data = NULL;
  EXULine *Target_Data = NULL;

  double        Input_Value = 0.00;
  unsigned char Valid_Input = 1, 
                Conversion_Displayed = 0,
		Source_Displayed     = 0,
		Target_Displayed     = 0;
		
  
char OldValue[20] = ""; /* keep track of string in edit_value */

void clean_exit(GtkWidget *Widget, GdkEvent *event, gpointer data)
{
  /* this is just a prolog to gtk_exit */
  delete_table(&Current_Table);
  gtk_exit(0);
}

GList *fill_menu(GList *liste)
{
  EXUList *cursor = Current_Table;
  
  while(cursor) 
  {
    liste = g_list_append(liste, cursor->data->iso4217);
    cursor=cursor->next;
  }
  return liste;
}

void update_display()
{
  char rate_str[255]; /* should be enough for all the stuff */
  char update_status_bar = 0;

  if(Target_Data && !Target_Displayed)
  {
    gtk_label_set_text(GTK_LABEL(target_name),Target_Data->cur_label);
    sprintf(rate_str,"%f",Target_Data->rate_i);
    gtk_label_set_text(GTK_LABEL(target_rate),rate_str);
    if(Source_Data) update_status_bar = 1;
    Target_Displayed = 1;
  }

  if(Source_Data && !Source_Displayed)
  {
    gtk_label_set_text(GTK_LABEL(source_name),Source_Data->cur_label);
    sprintf(rate_str,"%f",Source_Data->rate_i);
    gtk_label_set_text(GTK_LABEL(source_rate),rate_str);
    if(Target_Data) update_status_bar = 1;
    Source_Displayed = 1;
  }

  if(update_status_bar)
  {
    sprintf(rate_str,"Conversion rate from %s to %s is %f",
            Source_Data->iso4217, Target_Data->iso4217,
	    Target_Data->rate_i / Source_Data->rate_i);
    gtk_statusbar_pop(GTK_STATUSBAR(status_bar),status_bar_context);
    gtk_statusbar_push(GTK_STATUSBAR(status_bar),status_bar_context,rate_str);
  }
  
  if(!Conversion_Displayed)
  {
    if(Valid_Input)
    {
      sprintf(rate_str,"%f",(Input_Value * Target_Data->rate_i) / Source_Data->rate_i);
      gtk_label_set_text(GTK_LABEL(result_label),rate_str);
    }
    else gtk_label_set_text(GTK_LABEL(result_label),"invalid input");
    Conversion_Displayed = 1;
  }
}

unsigned char is_valid_number(char *input)
{
  unsigned char ret = 1, fpoint = 0;
  char *number = input;
  
  while(*number && ret)
  {
    if(*number<'0' || *number > '9')
    {
      if(*number != '.') ret = 0; 
      else if((++fpoint > 1) || /* do not accept more than 1 decimal point */
              (number == input)) /* do not accept floating point in first position */
	      ret = 0;
    }
    number++;
  }
  return ret;
}

/************************************************************************/
/* signal handlers                                                      */
/************************************************************************/

/*                                                                           on_entry2_changed */
void on_entry2_changed(GtkEditable *editable, gpointer user_data)
{
  char *text = gtk_entry_get_text(GTK_ENTRY(editable));
  
  if((Valid_Input = is_valid_number(text)) != 0) Input_Value = atof(text);
  Conversion_Displayed = 0;
  update_display();
}

/*                                                                     on_source_selection_get */
void on_source_selection_get(GtkWidget *widget, GtkSelectionData *data,
                             guint info, guint time, gpointer user_data)
{
  Source_Data = find_by_iso4217(gtk_entry_get_text(GTK_ENTRY(widget)));
  Source_Displayed = Conversion_Displayed = 0;
  update_display();
}

/*                                                                     on_target_selection_get */
void on_target_selection_get(GtkWidget *widget, GtkSelectionData *data,
                             guint info, guint time, gpointer user_data)
{
  Target_Data = find_by_iso4217(gtk_entry_get_text(GTK_ENTRY(widget)));
  Target_Displayed = Conversion_Displayed = 0;
  update_display();
}

/************************************************************************/
/* main window                                                          */
/************************************************************************/
GtkWidget *make_main_window()
{
  GtkWidget *main_window;
  GtkWidget *main_canvas;
  GtkWidget *target_frame;
  GtkWidget *target_canvas;
  GtkWidget *target_list;
  GtkWidget *target_combo;
  GtkWidget *target_rate_frame;
  GtkWidget *target_name_frame;
  GtkWidget *conversion_frame;
  GtkWidget *conversion_canvas;
  GtkWidget *result_frame;
  GtkWidget *close_button;
  GtkWidget *about_button;
  GtkWidget *source_frame;
  GtkWidget *source_canvas;
  GtkWidget *source_rate_frame;
  GtkWidget *source_name_frame;
  GtkWidget *source_list;
  GtkWidget *source_combo;
  GtkWidget *value_edit;

  GList *list_items = NULL;

/*                                                                                 Main Window */
    main_window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
    gtk_widget_set_name (main_window, "main_window");
    gtk_object_set_data (GTK_OBJECT (main_window), "main_window", main_window);
    gtk_window_set_title (GTK_WINDOW (main_window), "Currency Converter");
    gtk_window_set_policy (GTK_WINDOW (main_window), FALSE, FALSE, FALSE);

/*                                                                                 Main Canvas */
    main_canvas = gtk_fixed_new ();
    gtk_widget_set_name (main_canvas, "main_canvas");
    gtk_widget_ref (main_canvas);
    gtk_object_set_data_full (GTK_OBJECT (main_window), "main_canvas", main_canvas,
                              (GtkDestroyNotify) gtk_widget_unref);
    gtk_widget_show (main_canvas);
    gtk_container_add (GTK_CONTAINER (main_window), main_canvas);

/***********************************************************************************************/
/* Source data                                                                                 */
/***********************************************************************************************/

/*                                                                                Source Frame */
    source_frame = gtk_frame_new ("Source");
    gtk_widget_set_name (source_frame, "source_frame");
    gtk_widget_ref (source_frame);
    gtk_object_set_data_full (GTK_OBJECT (main_window), "source_frame", source_frame,
                              (GtkDestroyNotify) gtk_widget_unref);
    gtk_widget_show (source_frame);
    gtk_fixed_put (GTK_FIXED (main_canvas), source_frame, 0, 0);
    gtk_widget_set_uposition (source_frame, 0, 0);
    gtk_widget_set_usize (source_frame, 376, 48);
    gtk_frame_set_shadow_type (GTK_FRAME (source_frame), GTK_SHADOW_OUT);

/*                                                                               Source Canvas */
    source_canvas = gtk_fixed_new ();
    gtk_widget_set_name (source_canvas, "source_canvas");
    gtk_widget_ref (source_canvas);
    gtk_object_set_data_full (GTK_OBJECT (main_window), "source_canvas", source_canvas,
                              (GtkDestroyNotify) gtk_widget_unref);
    gtk_widget_show (source_canvas);
    gtk_container_add (GTK_CONTAINER (source_frame), source_canvas);

/*                                                                                 Source List */
    source_list = gtk_combo_new ();
    gtk_widget_set_name (source_list, "source_list");
    gtk_widget_ref (source_list);
    gtk_object_set_data_full (GTK_OBJECT (main_window), "source_list", source_list,
                              (GtkDestroyNotify) gtk_widget_unref);
    gtk_widget_show (source_list);
    gtk_fixed_put (GTK_FIXED (source_canvas), source_list, 0, 0);
    gtk_widget_set_uposition (source_list, 0, 0);
    gtk_widget_set_usize (source_list, 64, 24);
    
    list_items = fill_menu(list_items);

    gtk_combo_set_popdown_strings(GTK_COMBO(source_list), list_items);
    g_list_free (list_items);
    list_items = NULL;
/*                                                                                Source Combo */
    source_combo = GTK_COMBO (source_list)->entry;
    gtk_widget_set_name (source_combo, "source_combo");
    gtk_widget_ref (source_combo);
    gtk_object_set_data_full (GTK_OBJECT (main_window), "source_combo", source_combo,
                              (GtkDestroyNotify) gtk_widget_unref);
    gtk_widget_show (GTK_WIDGET(GTK_ENTRY(source_combo)));
    gtk_entry_set_editable (GTK_ENTRY (source_combo), FALSE);
    gtk_entry_set_text (GTK_ENTRY (source_combo), Source_Data->iso4217);

/*                                                                           Source Name Frame */
    source_name_frame = gtk_frame_new (NULL);
    gtk_widget_set_name (source_name_frame, "source_name_frame");
    gtk_widget_ref (source_name_frame);
    gtk_object_set_data_full (GTK_OBJECT (main_window), "source_name_frame", source_name_frame,
                              (GtkDestroyNotify) gtk_widget_unref);
    gtk_widget_show (source_name_frame);
    gtk_fixed_put (GTK_FIXED (source_canvas), source_name_frame, 72, 0);
    gtk_widget_set_uposition (source_name_frame, 72, 0);
    gtk_widget_set_usize (source_name_frame, 200, 24);
    gtk_frame_set_shadow_type (GTK_FRAME (source_name_frame), GTK_SHADOW_IN);

/*                                                                                 Source Name */
    source_name = gtk_label_new ("");
    gtk_widget_set_name (GTK_WIDGET(GTK_LABEL(source_name)), "source_name");
    gtk_widget_ref (GTK_WIDGET(GTK_LABEL(source_name)));
    gtk_object_set_data_full (GTK_OBJECT (main_window), "source_name", source_name,
                              (GtkDestroyNotify) gtk_widget_unref);
    gtk_widget_show (source_name);
    gtk_container_add (GTK_CONTAINER (source_name_frame), source_name);

/*                                                                           Source Rate Frame */
    source_rate_frame = gtk_frame_new (NULL);
    gtk_widget_set_name (source_rate_frame, "source_rate_frame");
    gtk_widget_ref (source_rate_frame);
    gtk_object_set_data_full (GTK_OBJECT (main_window), "source_rate_frame", source_rate_frame,
                              (GtkDestroyNotify) gtk_widget_unref);
    gtk_widget_show (source_rate_frame);
    gtk_fixed_put (GTK_FIXED (source_canvas), source_rate_frame, 276, 0);
    gtk_widget_set_uposition (source_rate_frame, 276, 0);
    gtk_widget_set_usize (source_rate_frame, 88, 24);
    gtk_frame_set_shadow_type (GTK_FRAME (source_rate_frame), GTK_SHADOW_IN);

/*                                                                                 Source Rate */
    source_rate = gtk_label_new ("");
    gtk_widget_set_name (GTK_WIDGET(GTK_LABEL(source_rate)), "source_rate");
    gtk_widget_ref (GTK_WIDGET(GTK_LABEL(source_rate)));
    gtk_object_set_data_full(GTK_OBJECT(GTK_WINDOW(main_window)), "source_rate",
                             GTK_WIDGET(GTK_LABEL(source_rate)),
                              (GtkDestroyNotify) gtk_widget_unref);
    gtk_widget_show (GTK_WIDGET(GTK_LABEL(source_rate)));
    gtk_container_add (GTK_CONTAINER(source_rate_frame), GTK_WIDGET(GTK_LABEL(source_rate)));
    gtk_label_set_justify (GTK_LABEL(source_rate), GTK_JUSTIFY_RIGHT);

/***********************************************************************************************/
/* Target data                                                                                 */
/***********************************************************************************************/

/*                                                                                Target Frame */
    target_frame = gtk_frame_new ("Target");
    gtk_widget_set_name (target_frame, "target_frame");
    gtk_widget_ref (target_frame);
    gtk_object_set_data_full (GTK_OBJECT (main_window), "target_frame", target_frame,
                              (GtkDestroyNotify) gtk_widget_unref);
    gtk_widget_show (target_frame);
    gtk_fixed_put (GTK_FIXED (main_canvas), target_frame, 0, 48);
    gtk_widget_set_uposition (target_frame, 0, 48);
    gtk_widget_set_usize (target_frame, 376, 47);
    gtk_frame_set_shadow_type (GTK_FRAME (target_frame), GTK_SHADOW_OUT);

/*                                                                               Target Canvas */
    target_canvas = gtk_fixed_new ();
    gtk_widget_set_name (target_canvas, "target_canvas");
    gtk_widget_ref (target_canvas);
    gtk_object_set_data_full (GTK_OBJECT (main_window), "target_canvas", target_canvas,
                              (GtkDestroyNotify) gtk_widget_unref);
    gtk_widget_show (target_canvas);
    gtk_container_add (GTK_CONTAINER (target_frame), target_canvas);

/*                                                                                 Target list */
    target_list = gtk_combo_new();
    gtk_widget_set_name(target_list, "target_list");
    gtk_widget_ref(target_list);
    gtk_object_set_data_full (GTK_OBJECT (main_window), "target_list", target_list,
                              (GtkDestroyNotify) gtk_widget_unref);
    gtk_widget_show(GTK_WIDGET(GTK_COMBO(target_list)));
    gtk_fixed_put(GTK_FIXED(target_canvas), target_list, 0, 0);
    gtk_widget_set_uposition (target_list, 0, 0);
    gtk_widget_set_usize(target_list, 64, 24);
  
    list_items = fill_menu(list_items);
  
    gtk_combo_set_popdown_strings (GTK_COMBO (target_list), list_items);
    g_list_free (list_items);
  list_items = NULL;
/*                                                                                Target Combo */
    target_combo = GTK_COMBO (target_list)->entry;
    gtk_widget_set_name (target_combo, "target_combo");
    gtk_widget_ref (target_combo);
    gtk_object_set_data_full (GTK_OBJECT (main_window), "target_combo", target_combo,
                              (GtkDestroyNotify) gtk_widget_unref);
    gtk_widget_show (target_combo);
    gtk_entry_set_editable (GTK_ENTRY (target_combo), FALSE);
    gtk_entry_set_text (GTK_ENTRY (target_combo), Current_Table->data->iso4217);

/*                                                                           Target Name Frame */
    target_name_frame = gtk_frame_new (NULL);
    gtk_widget_set_name (target_name_frame, "target_name_frame");
    gtk_widget_ref (target_name_frame);
    gtk_object_set_data_full (GTK_OBJECT (main_window), "target_name_frame", target_name_frame,
                              (GtkDestroyNotify) gtk_widget_unref);
    gtk_widget_show (target_name_frame);
    gtk_fixed_put (GTK_FIXED (target_canvas), target_name_frame, 72, 0);
    gtk_widget_set_uposition (target_name_frame, 72, 0);
    gtk_widget_set_usize (target_name_frame, 200, 24);
    gtk_frame_set_shadow_type (GTK_FRAME (target_name_frame), GTK_SHADOW_IN);

/*                                                                                Target Name */
    target_name = gtk_label_new ("");
    gtk_widget_set_name (GTK_WIDGET(GTK_LABEL(target_name)), "target_name");
    gtk_widget_ref (GTK_WIDGET(GTK_LABEL(target_name)));
    gtk_object_set_data_full (GTK_OBJECT (main_window), "target_name",
                              GTK_WIDGET(GTK_LABEL(target_name)),
                              (GtkDestroyNotify)gtk_widget_unref);
    gtk_widget_show (GTK_WIDGET(GTK_LABEL(target_name)));
    gtk_container_add (GTK_CONTAINER (target_name_frame), GTK_WIDGET(GTK_LABEL(target_name)));

/*                                                                           Target Rate Frame */
    target_rate_frame = gtk_frame_new (NULL);
    gtk_widget_set_name (target_rate_frame, "target_rate_frame");
    gtk_widget_ref (target_rate_frame);
    gtk_object_set_data_full (GTK_OBJECT (main_window), "target_rate_frame", target_rate_frame,
                              (GtkDestroyNotify) gtk_widget_unref);
    gtk_widget_show (target_rate_frame);
    gtk_fixed_put (GTK_FIXED (target_canvas), target_rate_frame, 276, 0);
    gtk_widget_set_uposition (target_rate_frame, 276, 0);
    gtk_widget_set_usize (target_rate_frame, 88, 24);
    gtk_frame_set_shadow_type (GTK_FRAME (target_rate_frame), GTK_SHADOW_IN);

/*                                                                                 Target Rate */
    target_rate = gtk_label_new ("");
    gtk_widget_set_name (target_rate, "target_rate");
    gtk_widget_ref (target_rate);
    gtk_object_set_data_full (GTK_OBJECT (main_window), "target_rate", target_rate,
                              (GtkDestroyNotify) gtk_widget_unref);
    gtk_widget_show (target_rate);
    gtk_container_add (GTK_CONTAINER (target_rate_frame), target_rate);
    gtk_label_set_justify (GTK_LABEL (target_rate), GTK_JUSTIFY_RIGHT);

/***********************************************************************************************/
/* Conversion Area                                                                             */
/***********************************************************************************************/

/*                                                                            Conversion Frame */
    conversion_frame = gtk_frame_new (NULL);
    gtk_widget_set_name (conversion_frame, "conversion_frame");
    gtk_widget_ref (conversion_frame);
    gtk_object_set_data_full (GTK_OBJECT (main_window), "conversion_frame", conversion_frame,
                              (GtkDestroyNotify) gtk_widget_unref);
    gtk_widget_show (conversion_frame);
    gtk_fixed_put (GTK_FIXED (main_canvas), conversion_frame, 0, 104);
    gtk_widget_set_uposition (conversion_frame, 0, 104);
    gtk_widget_set_usize (conversion_frame, 372, 36);
    gtk_frame_set_shadow_type (GTK_FRAME (conversion_frame), GTK_SHADOW_OUT);

/*                                                                           Conversion Canvas */
    conversion_canvas = gtk_fixed_new ();
    gtk_widget_set_name (conversion_canvas, "conversion_canvas");
    gtk_widget_ref (conversion_canvas);
    gtk_object_set_data_full (GTK_OBJECT (main_window), "conversion_canvas", conversion_canvas,
                              (GtkDestroyNotify) gtk_widget_unref);
    gtk_widget_show (conversion_canvas);
    gtk_container_add (GTK_CONTAINER (conversion_frame), conversion_canvas);

/*                                                                                Value Edit   */
    value_edit = gtk_entry_new_with_max_length(10);
    gtk_widget_set_name (value_edit, "value_edit");
    gtk_widget_ref (value_edit);
    gtk_object_set_data_full (GTK_OBJECT (main_window), "value_edit", value_edit,
                              (GtkDestroyNotify) gtk_widget_unref);
    gtk_widget_show (value_edit);
    gtk_fixed_put (GTK_FIXED (conversion_canvas), value_edit, 0, 4);
    gtk_widget_set_uposition (value_edit, 0, 4);
    gtk_widget_set_usize (value_edit, 140, 24);
    GTK_WIDGET_SET_FLAGS (value_edit, GTK_CAN_DEFAULT);
    gtk_widget_set_events (value_edit, GDK_KEY_PRESS_MASK);
    gtk_widget_set_extension_events (value_edit, GDK_EXTENSION_EVENTS_ALL);

/*                                                                                Result Frame */
    result_frame = gtk_frame_new (NULL);
    gtk_widget_set_name (result_frame, "result_frame");
    gtk_widget_ref (result_frame);
    gtk_object_set_data_full (GTK_OBJECT (main_window), "result_frame", result_frame,
                              (GtkDestroyNotify) gtk_widget_unref);
    gtk_widget_show (result_frame);
    gtk_fixed_put (GTK_FIXED (conversion_canvas), result_frame, 152, 4);
    gtk_widget_set_uposition (result_frame, 152, 4);
    gtk_widget_set_usize (result_frame, 212, 24);
    gtk_frame_set_shadow_type (GTK_FRAME (result_frame), GTK_SHADOW_IN);

/*                                                                                Result Label */
    result_label = gtk_label_new ("");
    gtk_widget_set_name (result_label, "result_label");
    gtk_widget_ref (result_label);
    gtk_object_set_data_full (GTK_OBJECT (main_window), "result_label", result_label,
                              (GtkDestroyNotify) gtk_widget_unref);
    gtk_widget_show (result_label);
    gtk_container_add (GTK_CONTAINER (result_frame), result_label);
    gtk_label_set_justify (GTK_LABEL (result_label), GTK_JUSTIFY_RIGHT);

/***********************************************************************************************/
/* Buttons                                                                                     */
/***********************************************************************************************/

/*                                                                                About Button */
    about_button = gtk_button_new_with_label ("About");
    gtk_widget_set_name (about_button, "about_button");
    gtk_widget_ref (about_button);
    gtk_object_set_data_full (GTK_OBJECT (main_window), "about_button", about_button,
                              (GtkDestroyNotify) gtk_widget_unref);
    gtk_widget_show (about_button);
    gtk_fixed_put (GTK_FIXED (main_canvas), about_button, 88, 144);
    gtk_widget_set_uposition (about_button, 88, 144);
    gtk_widget_set_usize (about_button, 140, 24);

/*                                                                                Close Button */
    close_button = gtk_button_new_with_label ("Close");
    gtk_widget_set_name (close_button, "close_button");
    gtk_widget_ref (close_button);
    gtk_object_set_data_full (GTK_OBJECT(main_window), "close_button", close_button,
                              (GtkDestroyNotify) gtk_widget_unref);
    gtk_widget_show (close_button);
    gtk_fixed_put (GTK_FIXED(main_canvas), close_button, 232, 144);
    gtk_widget_set_uposition (close_button, 232, 144);
    gtk_widget_set_usize (close_button, 140, 24);
    
/*                                                                                  Status Bar */
    status_bar = gtk_statusbar_new ();
    gtk_widget_set_name (status_bar, "status_bar");
    gtk_widget_ref (status_bar);
    gtk_object_set_data_full (GTK_OBJECT(GTK_WINDOW(main_window)), "status_bar",
                              GTK_WIDGET(GTK_STATUSBAR(status_bar)),
                              (GtkDestroyNotify)gtk_widget_unref);
    gtk_widget_show (status_bar);
    gtk_fixed_put (GTK_FIXED (main_canvas), status_bar, 0, 172);
    gtk_widget_set_uposition (status_bar, 0, 172);
    gtk_widget_set_usize (status_bar, 376, 24);
    gtk_container_set_border_width (GTK_CONTAINER (status_bar), 2);
    gtk_widget_set_sensitive (status_bar, FALSE);
    
    status_bar_context = gtk_statusbar_get_context_id(GTK_STATUSBAR(status_bar),"status_bar");
    gtk_statusbar_push(GTK_STATUSBAR(status_bar),status_bar_context,"Initializing...");

    gtk_widget_grab_focus(value_edit);
    gtk_widget_grab_default(value_edit);


    gtk_signal_connect_after(GTK_OBJECT(GTK_ENTRY(source_combo)),"changed",
                             GTK_SIGNAL_FUNC(on_source_selection_get),NULL);

    gtk_signal_connect_after(GTK_OBJECT(GTK_ENTRY(target_combo)), "changed",
                             GTK_SIGNAL_FUNC(on_target_selection_get), NULL);

    gtk_signal_connect (GTK_OBJECT(GTK_ENTRY(value_edit)), "changed",
                        GTK_SIGNAL_FUNC(on_entry2_changed), NULL);

    gtk_signal_connect (GTK_OBJECT(about_button), "clicked", (GtkSignalFunc)AboutDialog, NULL);

/* exit cleanly */
    gtk_signal_connect(GTK_OBJECT(close_button), "clicked"    , (GtkSignalFunc)clean_exit, NULL);
    gtk_signal_connect(GTK_OBJECT(main_window), "delete_event", (GtkSignalFunc)clean_exit, NULL);
    gtk_signal_connect(GTK_OBJECT(main_window), "destroy"     , (GtkSignalFunc)clean_exit, NULL);

    update_display();

    return main_window;
}

/************************************************************************/
/* main()                                                               */
/************************************************************************/

int main(int argc, char *argv[])
{
  int ret = 0;
  if(load_table(DATA_FILE,&Current_Table))
  {
    gtk_set_locale ();
    gtk_init (&argc, &argv);
    Source_Data = Target_Data = Current_Table->data;
    gtk_widget_show(GTK_WIDGET(GTK_WINDOW(make_main_window())));
    gtk_main ();
  }
  else puts("problem opening file.");
  return ret;
}
